/******** <lalVerbatim file="BlockConditionDataCV"> ********
Author: Tibbits, M. M.
$Id: BlockConditionData.c,v 1.2 2002/07/05 20:52:22 tibbits Exp $
********* </lalVerbatim> ********/

/******************************* <lalLaTeX file="BlockConditionDataC">
\subsection{Module \texttt{BlockConditionData.c}}
\label{s:BlockConditionData.c}

Routine that processes the input data object and repacks it into the
block parameter structure.  It then also sends out the first vector
to be processed.

\subsubsection*{Prototypes}
\input{BlockConditionDataCP}
\input{BlockDistributeDataMasterCP}
\input{BlockDistributeDataSlaveCP}
\input{BlockParseDataCP}
\index{\texttt{LALConditionData()}}

\subsection*{Description}

Within the \texttt{LALConditionData} Routine, the block search master sends
the input object to the \texttt{BlockParseData} function.  This function
repacks the input in the masters parameter structure and stores the input
data in the parameter's data member.  The Master then sends the first vector
to each node to begin processing.  On each slave node, each slave simply
receives the first data vector and then exits.

\subsection*{Algorithms}

At this time, the \texttt{LALConditionData} routine does not manipulate
the data in any way.  It merely repacks the data in a LAL Datatype.

\subsection*{Uses}
 
\begin{itemize}
\item \texttt{BlockDistributeDataMaster}
\item \texttt{BlockDistributeDataSlave}
\item \texttt{LALMPISendMsg}
\item \texttt{LALMPIRecvMsg}
\item \texttt{LALMPISendINT4}
\item \texttt{LALMPIRecvINT4}
\item \texttt{LALMPISendREAL8}
\item \texttt{LALMPIRecvREAL8}
\item \texttt{LALDCreateVector}
\item \texttt{LALMPISendREAL8Vector}
\item \texttt{LALMPIRecvREAL8Vector}
\item \texttt{BlockParseData}
\item \texttt{LALDCreateVectorSequence}
\item \texttt{BlockBuildOutput} 
\item \texttt{LALDDestroyVector}
\end{itemize}

\subsection*{Notes}

\vfill{\footnotesize\input{BlockConditionDataCV}}

 *************************************************** </lalLaTeX> */


#include "BlockConditionData.h"
#include "BlockApplySearch.h"
#include <lal/SeqFactories.h>

NRCSID( BLOCKCONDITIONDATAC, "$Id: BlockConditionData.c,v 1.2 2002/07/05 20:52:22 tibbits Exp $");

/************************ <lalVerbatim file="BlockConditionDataCP"> */
void LALConditionData
(
	LALStatus	*status,
	LALSearchInput	*inout,
	void		*searchParams,
	LALMPIParams	*mpiParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*myParams;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "ConditionData", BLOCKCONDITIONDATAC );
	ATTATCHSTATUSPTR (status);

	/*  Sanity Checks on Input  */
	ASSERT( inout, status, BLOCKCONDITIONDATAH_ENULL, BLOCKCONDITIONDATAH_MSGENULL);

        ASSERT(searchParams, status,BLOCKCONDITIONDATAH_ENULL, BLOCKCONDITIONDATAH_MSGENULL);

	/*  Variable Initialization  */
	myParams = ((BlockSearchParams*)(searchParams));

	/*  Get rank  */
	MPI_Comm_rank(*(mpiParams->comm), &(myParams->rank));

	/*  Start Master/Slave version  */
	/*  Which node is this running on?  */
	if (myParams->searchMaster)
	{
		/* master */
		BlockDistributeDataMaster(status->statusPtr, inout, searchParams, mpiParams );
	}
	else
	{
		/* slave */
		BlockDistributeDataSlave(status->statusPtr, inout, searchParams, mpiParams );
	}

	DETATCHSTATUSPTR( status );
	RETURN( status );
}


/*
 * Function BlockDistributeDataMaster
 *
 * Called from LALConditionData only on the search master node. It
 * uses the BlockParseData routine to repack the input and then it
 * sends the a data vector to each waiting slave.
 */

/************************ <lalVerbatim file="BlockDistributeDataMasterCP"> */
void BlockDistributeDataMaster
(
	LALStatus	*status,
	LALSearchInput	*inout,
	void		*searchParams,
	LALMPIParams	*mpiParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*myParams;
	REAL8Vector		*tempVector;
	MPIMessage		mpiMsg;
	MPI_Comm		*comm;
	UINT4			currSlaves;
	UINT4			iterator;
	UINT4			segLength;
	UINT4			segLimit;
	INT4			id;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "DistributeDataMaster", BLOCKCONDITIONDATAC );
	ATTATCHSTATUSPTR (status);

	/*  No input checking done here  */
	/*  All assertions in LALConditionData Routine above  */

	/*  Variable Initialization  */
	comm		= ((MPI_Comm*)(mpiParams->comm));
	myParams	= ((BlockSearchParams*)(searchParams));

	/* get the number of search slaves */
	MPI_Comm_size(*(mpiParams->comm), &(myParams->numSlaves));

	/* The number returned INCLUDES the search master, so subtract 1 */
	myParams->numSlaves -= 1;

	currSlaves	= myParams->numSlaves;
	id		= myParams->nextVectorToSend = 0;
	iterator	= 0;
	segLimit	= 0;
	tempVector	= NULL;

	/*  Repack data into usable LAL Structure  */
	BlockParseData ( status->statusPtr, inout, myParams );

        /* Create vector to store Max Value of Data Sequence */
        LALDCreateVector(status->statusPtr, &(myParams->value), myParams->numberOfSequences);
	CHECKSTATUSPTR( status );

        /* Create vector to store index of max value */
        LALI4CreateVector(status->statusPtr, &(myParams->ndx), myParams->numberOfSequences);
	CHECKSTATUSPTR( status );

        /* Create vector to store Rho2 results */
        LALDCreateVector(status->statusPtr, &(myParams->res), myParams->numberOfSequences);
	CHECKSTATUSPTR( status );

	/*  While there are still Slaves without Data, send more  */
	while(currSlaves > 0 )
	{
		/*  Wait for Slaves to say, "I'm here"  */
		LALMPIRecvMsg( status->statusPtr, &mpiMsg, (*comm));
		CHECKSTATUSPTR( status );

		if ( id != -1 )
			segLength = myParams->data->vector[id]->length;

		/*  Send length of vector  */
		LALMPISendUINT4( status->statusPtr, &(segLength), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send index of data in vector sequence  */
                LALMPISendINT4( status->statusPtr, &(id), mpiMsg.source, (*comm));
                CHECKSTATUSPTR( status );

		if( id != -1 )
		{
			/*  Create and fill temporary vector to send to slave  */
			LALDCreateVector(status->statusPtr, &tempVector, segLength );
			CHECKSTATUSPTR( status );

			for( iterator = 0; iterator < tempVector->length; iterator++)
			{
				tempVector->data[iterator] = myParams->data->vector[id]->data[iterator];
			}

			/*  Send Data to Slave  */
			LALMPISendREAL8Vector(status->statusPtr, tempVector, mpiMsg.source, (*comm));
			CHECKSTATUSPTR( status );

			/*  Clean up temporary Mem each time used  */
			LALDDestroyVector(status->statusPtr, &tempVector);
			CHECKSTATUSPTR( status );
			tempVector	= NULL;
		}

		currSlaves--;
		id++;

		if ( id == myParams->numberOfSequences )
			id = -1;
	}

	/*  Set parameter to id of the Next Vector to Send  */
	myParams->nextVectorToSend = id;

	/*  Normal Exit  */
	DETATCHSTATUSPTR( status );
	RETURN( status );

}

/*
 * Function BlockDistributeDataSlave
 *
 * Called from LALConditionData on each slave node. It
 * waits for the master to send the first data vector for
 * it to process and then it exits.
 */

/************************ <lalVerbatim file="BlockDistributeDataSlaveCP"> */
void BlockDistributeDataSlave
(
	LALStatus	*status,
	LALSearchInput	*inout,
	void		*searchParams,
	LALMPIParams	*mpiParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*myParams;
	REAL8Vector		*local;
	MPIMessage		sendMsg;
	MPI_Comm		*comm;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "DistributeDataSlave", BLOCKCONDITIONDATAC );
	ATTATCHSTATUSPTR (status);

	/*  Variable Initialization  */
	myParams	= ((BlockSearchParams*)(searchParams));
	comm		= ((MPI_Comm*)(mpiParams->comm));
	local		= NULL;
	sendMsg.source	= myParams->rank;
	sendMsg.msg	= PassData;
	sendMsg.send	= 1;

	/* Create length 1 array to transmit/receive the result */
	LALDCreateVector(status->statusPtr, &(myParams->res), 1);

	/*  Send Message to say I am ready  */
	LALMPISendMsg( status->statusPtr, &sendMsg, 0, (*comm));
	CHECKSTATUSPTR( status );	

	/*  How much Data?  */
	LALMPIRecvUINT4( status->statusPtr, &(myParams->segmentLength), 0, (*comm));
	CHECKSTATUSPTR (status);

	/*  Receive the index of the vector  */
        LALMPIRecvINT4( status->statusPtr, &(myParams->vectorId), 0, (*comm));
        CHECKSTATUSPTR (status);

	if( myParams->vectorId != -1 )
	{
		/*  Create new structure to hold local data  */
		LALDCreateVector(status->statusPtr, &local, myParams->segmentLength );
		CHECKSTATUSPTR (status);

		/*  Receive input data  */
		LALMPIRecvREAL8Vector(status->statusPtr, local, 0, (*comm));
		CHECKSTATUSPTR (status);

		myParams->vec = local;
	}

	/*  Normal Exit  */
	DETATCHSTATUSPTR( status );
	RETURN( status );
}

/************************ <lalVerbatim file="BlockParseDataCP"> */
void BlockParseData
(
	LALStatus	*status,
	LALSearchInput	*inout,
	void		*searchParams
)
/*********************************************** </lalVerbatim> */
{

	/*  Variable Declarations  */
	BlockSearchParams	*params;
	REAL8VectorVector	*tempVector;
	BOOLEAN			dataFound;
	UINT4			iterator;
	UINT4			segmentCount;
	UINT4			correctedNumberOfSequences;
	UINT4			numberOfFullSequences;
	UINT4			sequenceNumber;
	UINT4			offset;
	INT4			length;
	INT4			count;
	INT4			index;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "BlockParseData", BLOCKCONDITIONDATAC );
	ATTATCHSTATUSPTR (status);

	/*  Variable Initialization  */
	params				= ((BlockSearchParams*)(searchParams));
	tempVector			= params->data = NULL;
	index				= 0;
	count				= 0;
	offset				= 0;
	length				= 0;
	dataFound			= 0;
	segmentCount			= 0;
	sequenceNumber			= 0;
	numberOfFullSequences		= 0;
	correctedNumberOfSequences	= 0;

	/*  Check that input exists  */
	ASSERT(inout, status, BLOCKCONDITIONDATAH_ENULL, BLOCKCONDITIONDATAH_MSGENULL);

	/*  Check that input exists  */
	ASSERT(inout->sequences, status, BLOCKCONDITIONDATAH_ENULL, BLOCKCONDITIONDATAH_MSGENULL);

	/*  Initialize the array of Pointers  */
	tempVector = (REAL8VectorVector*)(LALMalloc(sizeof(REAL8VectorVector)));
	if( !tempVector )
	{
		ABORT (status, BLOCKCONDITIONDATAH_EALOC, BLOCKCONDITIONDATAH_MSGEALOC);
	}

	for(iterator = 0; iterator < inout->numberSequences; iterator++ )
	{
		/*  Using these variables as temporaries - no naming significance here  */
		length = inout->sequences[iterator].dimensions[0] / params->maxVectorLength + 1;

		count += ( length );

		if ( (inout->sequences[iterator].dimensions[0] % params->maxVectorLength) == 0 )
			count--;
	}

	length = 0;

	params->numberOfSequences = tempVector->numberOfVectors = count;

	/*  Allocate space for each pointer in the array  */
	tempVector->vector = (REAL8Vector**) LALMalloc(sizeof(REAL8Vector*) * count);

	for(iterator = 0; iterator < inout->numberSequences; iterator++ )
	{
		index = 0;
		count = 0;

		while( index < inout->sequences[iterator].dimensions[0] )
		{
			if ( ( params->maxVectorLength + index ) < inout->sequences[iterator].dimensions[0] )
				length = params->maxVectorLength;
			else
				length = inout->sequences[iterator].dimensions[0] - index;

			index += length;

			if ( length < 6 )
				length = 6;

			tempVector->vector[sequenceNumber] = NULL;
			LALDCreateVector(status->statusPtr, &(tempVector->vector[sequenceNumber]), length);

			for (count = (index - length); count < index; count++ )
				tempVector->vector[sequenceNumber]->data[length - (index - count)] = ((REAL8)(inout->sequences[iterator].data.real4[count]));

			sequenceNumber++;
		}
	}

	/*  Keep the metadata intact to send to the database  */

	params->times.startSec = (INT8) inout->sequences[0].range.dTime.startSec;
	params->times.startNan = (INT8) inout->sequences[0].range.dTime.startNan;
	params->times.stopSec = (INT8) inout->sequences[0].range.dTime.stopSec;
	params->times.stopNan = (INT8) inout->sequences[0].range.dTime.stopNan;
	params->times.timeStepSize = (INT8) inout->sequences[0].range.dTime.timeStepSize;
	params->times.numberSamples = (INT8) inout->sequences[0].range.dTime.numberSamples;

	/*  This is not correct for production model  */
	/*  HARDCODED Channel Name  */
	strcpy(params->channel, "H2");

	/*  Store the newly created input structure in the master's parameter structure  */
	params->data = tempVector;

	/*  Normal Exit  */
	DETATCHSTATUSPTR (status);
	RETURN(status);
}
